header, footer {
  display: block;
}

body > header {
  margin-bottom: 1em;
  border-bottom: 1px solid #C0C0C0;
}

body > header > h1 {
  margin-bottom: 0.3em;
  font-size: 1.25em;
}

body > footer {
  margin-top: 2em;
  border-top: 1px solid #C0C0C0;
}

.breadcrumbs a {
  margin: 0 0.2em;
  text-decoration: none;
}

.breadcrumbs a[href]:hover {
  text-decoration: underline;
}

ul.breadcrumbs > li {
  display: inline;
}

.icon[data-type="link"] {
  background: url("link.png?a=static") center no-repeat;
}

.icon[data-type="file"] {
  background: url("item.png?a=static") center no-repeat;
}

.icon[data-type="dir"] {
  background: url("fclose.png?a=static") center no-repeat;
}

#data-table {
  border-collapse: collapse;
}

#panel {
  padding: 0.5em;
}

#panel fieldset {
  margin: 0;
  border: 0;
  padding: 0;
}
